# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for the store service."""

import datetime
import platform
from typing import cast
from unittest import mock

import craft_store
import craft_store.errors
import distro
import pytest
import requests
from craft_cli.pytest_plugin import RecordingEmitter
from craft_store import models, publisher
from hypothesis import given, strategies

import charmcraft
from charmcraft import application, errors
from charmcraft.models.project import CharmLib
from charmcraft.services.store import StoreService
from charmcraft.store import client
from tests import get_fake_revision


@pytest.fixture
def store(service_factory, mock_store_anonymous_client) -> StoreService:
    store = StoreService(app=application.APP_METADATA, services=service_factory)
    store.client = mock.Mock(spec_set=client.Client)
    store.anonymous_client = mock_store_anonymous_client
    return store


@pytest.fixture(scope="module")
def reusable_store():
    store = StoreService(app=application.APP_METADATA, services=None)
    store.client = mock.Mock(spec_set=craft_store.StoreClient)
    store._publisher = mock.Mock(spec_set=craft_store.PublisherGateway)
    return store


def test_user_agent(store):
    assert (
        store._user_agent
        == f"Charmcraft/{charmcraft.__version__} ({store._ua_system_info})"
    )


@pytest.mark.parametrize("system", ["Macos"])
@pytest.mark.parametrize("release", ["10", "11", "12"])
@pytest.mark.parametrize("machine", ["x86_64", "arm64", "riscv64"])
@pytest.mark.parametrize("python", ["CPython", "PyPy"])
@pytest.mark.parametrize("python_version", ["3.10", "3.12"])
def test_ua_system_info_non_linux(
    monkeypatch, store, system, release, machine, python, python_version
):
    monkeypatch.setattr(platform, "system", lambda: system)
    monkeypatch.setattr(platform, "release", lambda: release)
    monkeypatch.setattr(platform, "machine", lambda: machine)
    monkeypatch.setattr(platform, "python_implementation", lambda: python)
    monkeypatch.setattr(platform, "python_version", lambda: python_version)

    assert (
        store._ua_system_info
        == f"{system} {release}; {machine}; {python} {python_version}"
    )


@pytest.mark.parametrize("machine", ["x86_64", "arm64", "riscv64"])
@pytest.mark.parametrize("python", ["CPython", "PyPy"])
@pytest.mark.parametrize("python_version", ["3.10", "3.12"])
@pytest.mark.parametrize("distro_name", ["Ubuntu", "Debian", "Something"])
@pytest.mark.parametrize("distro_version", ["1", "24.04", "version"])
def test_ua_system_info_linux(
    monkeypatch, store, machine, python, python_version, distro_name, distro_version
):
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(platform, "release", lambda: "6.5.0")
    monkeypatch.setattr(platform, "machine", lambda: machine)
    monkeypatch.setattr(platform, "python_implementation", lambda: python)
    monkeypatch.setattr(platform, "python_version", lambda: python_version)
    monkeypatch.setattr(distro, "name", lambda: distro_name)
    monkeypatch.setattr(distro, "version", lambda: distro_version)

    assert (
        store._ua_system_info
        == f"Linux 6.5.0; {machine}; {python} {python_version}; {distro_name} {distro_version}"
    )


def test_setup_with_error(emitter: RecordingEmitter, store):
    store.ClientClass = mock.Mock(
        side_effect=[craft_store.errors.NoKeyringError, "I am a store!"]
    )

    store.setup()

    assert store.client == "I am a store!"
    emitter.assert_progress(
        "WARNING: Cannot get a keyring. Every store interaction that requires "
        "authentication will require you to log in again.",
        permanent=True,
    )


def test_get_description_default(monkeypatch, store):
    monkeypatch.setattr(platform, "node", lambda: "my-hostname")

    assert store._get_description() == "charmcraft@my-hostname"


@given(text=strategies.text())
def test_get_description_override(reusable_store, text):
    assert reusable_store._get_description(text) == text


@given(
    permissions=strategies.lists(strategies.text()),
    description=strategies.text(),
    ttl=strategies.integers(min_value=1),
    channels=strategies.lists(strategies.text()),
)
def test_login(reusable_store, permissions, description, ttl, channels):
    client = cast(mock.Mock, reusable_store.client)
    client.reset_mock(return_value=True, side_effect=True)

    reusable_store.login(
        permissions=permissions, description=description, ttl=ttl, channels=channels
    )

    client.login.assert_called_once_with(
        permissions=permissions,
        description=description,
        ttl=ttl,
        packages=None,
        channels=channels,
    )


def test_login_failure(store):
    client = cast(mock.Mock, store.client)
    client.login.side_effect = craft_store.errors.CredentialsAlreadyAvailable(
        "charmcraft", "host"
    )

    with pytest.raises(
        errors.CraftError, match="Cannot login because credentials were found"
    ):
        store.login()


def test_logout(store):
    client = cast(mock.Mock, store.client)

    store.logout()

    client.logout.assert_called_once_with()


def test_create_tracks(reusable_store: StoreService):
    mock_create = cast(mock.Mock, reusable_store._publisher.create_tracks)
    mock_md = cast(mock.Mock, reusable_store._publisher.get_package_metadata)
    user_track: publisher.CreateTrackRequest = {
        "name": "my-track",
        "automatic-phasing-percentage": None,
    }
    created_at = {"created-at": datetime.datetime.now()}
    return_track = publisher.Track.unmarshal(user_track | created_at)
    mock_md.return_value = publisher.RegisteredName.unmarshal(
        {
            "id": "mentalism",
            "private": False,
            "publisher": {"id": "EliBosnick"},
            "status": "hungry",
            "store": "charmhub",
            "type": "charm",
            "tracks": [
                return_track,
                publisher.Track.unmarshal(
                    {
                        "name": "latest",
                        "automatic-phasing-percentage": None,
                    }
                    | created_at
                ),
            ],
        }
    )

    assert reusable_store.create_tracks("my-name", user_track) == [return_track]
    mock_create.assert_called_once_with("my-name", user_track)


@pytest.mark.parametrize(
    ("updates", "expected_request"),
    [
        pytest.param({}, [], id="empty"),
        pytest.param(
            {123: ["amd64", "riscv64"]},
            [
                models.CharmResourceRevisionUpdateRequest(
                    revision=123,
                    bases=[
                        models.RequestCharmResourceBase(
                            architectures=["amd64", "riscv64"]
                        )
                    ],
                )
            ],
        ),
        pytest.param(
            {
                123: ["amd64", "riscv64"],
                456: ["all"],
            },
            [
                models.CharmResourceRevisionUpdateRequest(
                    revision=123,
                    bases=[
                        models.RequestCharmResourceBase(
                            architectures=["amd64", "riscv64"]
                        )
                    ],
                ),
                models.CharmResourceRevisionUpdateRequest(
                    revision=456,
                    bases=[models.RequestCharmResourceBase(architectures=["all"])],
                ),
            ],
        ),
    ],
)
def test_set_resource_revisions_architectures_request_form(
    store, updates, expected_request
):
    store.client.list_resource_revisions.return_value = []

    store.set_resource_revisions_architectures("my-charm", "my-file", updates)

    store.client.update_resource_revisions.assert_called_once_with(
        *expected_request,
        name="my-charm",
        resource_name="my-file",
    )


@pytest.mark.parametrize(
    ("updates", "store_response", "expected"),
    [
        ({}, [], []),
        (
            {123: ["all"]},
            [
                get_fake_revision(
                    bases=[models.ResponseCharmResourceBase()], revision=0
                ),
                get_fake_revision(
                    bases=[models.ResponseCharmResourceBase()], revision=123
                ),
            ],
            [
                get_fake_revision(
                    bases=[models.ResponseCharmResourceBase()], revision=123
                )
            ],
        ),
    ],
)
def test_set_resource_revisions_architectures_response_form(
    store, updates, store_response, expected
):
    store.client.list_resource_revisions.return_value = store_response

    actual = store.set_resource_revisions_architectures("my-charm", "my-file", updates)

    assert actual == expected


def test_get_credentials(monkeypatch, store):
    mock_client_class = mock.Mock(spec_set=craft_store.StoreClient)
    mock_client = mock_client_class.return_value
    monkeypatch.setattr(craft_store, "StoreClient", mock_client_class)

    store.get_credentials()

    mock_client_class.assert_called_once_with(
        application_name="charmcraft",
        base_url=mock.ANY,
        storage_base_url=mock.ANY,
        endpoints=mock.ANY,
        environment_auth=None,
        user_agent=store._user_agent,
        ephemeral=True,
    )
    mock_client.login.assert_called_once_with(
        permissions=mock.ANY,
        description=store._get_description(),
        ttl=mock.ANY,
        packages=None,
        channels=None,
    )


@given(name=strategies.text())
def test_get_package_metadata(reusable_store: StoreService, name: str):
    mock_get = cast(mock.Mock, reusable_store._publisher.get_package_metadata)
    mock_get.reset_mock()  # Hypothesis runs this multiple times with the same fixture.

    reusable_store.get_package_metadata(name)

    mock_get.assert_called_once_with(name)


@pytest.mark.parametrize("requests", [[], [{}]])
def test_release(reusable_store: StoreService, requests):
    name = "my-charm"
    mock_release = cast(mock.Mock, reusable_store._publisher.release)
    mock_release.reset_mock()

    reusable_store.release(name, requests)

    mock_release.assert_called_once_with(name, requests=requests)


@pytest.mark.parametrize(
    ("store_response", "expected"),
    [
        pytest.param(
            publisher.Releases(
                channel_map=[], package=publisher.Package(channels=[]), revisions=[]
            ),
            [],
            id="empty",
        ),
        pytest.param(
            publisher.Releases(
                channel_map=[
                    publisher.ChannelMap(
                        base=publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        ),
                        channel="latest/edge",
                        revision=1,
                        when=datetime.datetime(2020, 1, 1),
                    )
                ],
                package=publisher.Package(channels=[]),
                revisions=[
                    publisher.CharmRevision(
                        revision=1,
                        bases=[
                            publisher.Base(
                                name="ubuntu", channel="25.10", architecture="riscv64"
                            )
                        ],
                        version="1",
                        status="peachy",
                        created_at=datetime.datetime(2020, 1, 1),
                        size=0,
                    )
                ],
            ),
            [
                {
                    "revision": 1,
                    "bases": [
                        publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        )
                    ],
                    "resources": [],
                    "version": "1",
                }
            ],
            id="basic",
        ),
        pytest.param(
            publisher.Releases(
                channel_map=[
                    publisher.ChannelMap(
                        base=publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        ),
                        channel="latest/edge",
                        revision=1,
                        when=datetime.datetime(2020, 1, 1),
                        resources=[
                            publisher.Resource(name="file", revision=2, type="file"),
                            publisher.Resource(
                                name="rock", revision=3, type="oci-image"
                            ),
                        ],
                    )
                ],
                package=publisher.Package(channels=[]),
                revisions=[
                    publisher.CharmRevision(
                        revision=1,
                        bases=[
                            publisher.Base(
                                name="ubuntu", channel="25.10", architecture="riscv64"
                            )
                        ],
                        version="1",
                        status="peachy",
                        created_at=datetime.datetime(2020, 1, 1),
                        size=0,
                    )
                ],
            ),
            [
                {
                    "revision": 1,
                    "bases": [
                        publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        )
                    ],
                    "resources": [
                        {"name": "file", "revision": 2},
                        {"name": "rock", "revision": 3},
                    ],
                    "version": "1",
                }
            ],
            id="resources",
        ),
        pytest.param(
            publisher.Releases(
                channel_map=[
                    publisher.ChannelMap(
                        base=publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        ),
                        channel="latest/edge",
                        revision=1,
                        when=datetime.datetime(2020, 1, 1),
                        resources=[
                            publisher.Resource(name="file", revision=2, type="file"),
                            publisher.Resource(
                                name="rock", revision=3, type="oci-image"
                            ),
                        ],
                    ),
                    publisher.ChannelMap(
                        base=publisher.Base(
                            name="ubuntu", channel="25.11", architecture="riscv64"
                        ),
                        channel="latest/edge",
                        revision=1,
                        when=datetime.datetime(2020, 1, 1),
                        resources=[
                            publisher.Resource(name="file", revision=2, type="file"),
                            publisher.Resource(
                                name="rock", revision=3, type="oci-image"
                            ),
                        ],
                    ),
                ],
                package=publisher.Package(channels=[]),
                revisions=[
                    publisher.CharmRevision(
                        revision=1,
                        bases=[
                            publisher.Base(
                                name="ubuntu", channel="25.10", architecture="riscv64"
                            ),
                            publisher.Base(
                                name="ubuntu", channel="25.11", architecture="riscv64"
                            ),
                        ],
                        version="1",
                        status="peachy",
                        created_at=datetime.datetime(2020, 1, 1),
                        size=0,
                    )
                ],
            ),
            [
                {
                    "revision": 1,
                    "bases": [
                        publisher.Base(
                            name="ubuntu", channel="25.10", architecture="riscv64"
                        ),
                        publisher.Base(
                            name="ubuntu", channel="25.11", architecture="riscv64"
                        ),
                    ],
                    "resources": [
                        {"name": "file", "revision": 2},
                        {"name": "rock", "revision": 3},
                    ],
                    "version": "1",
                }
            ],
            id="multiple-bases",
        ),
    ],
)
def test_get_revisions_on_channel(
    reusable_store: StoreService, store_response, expected
):
    name = "my-charm"
    channel = "latest/edge"
    mock_list = cast(mock.Mock, reusable_store._publisher.list_releases)
    mock_list.return_value = store_response

    actual = reusable_store.get_revisions_on_channel(name, channel)

    assert actual == expected


@pytest.mark.parametrize(
    ("channel", "candidates", "expected"),
    [
        ("latest/stable", [], []),
        (
            "latest/edge",
            [{"revision": 1, "resources": []}],
            [{"channel": "latest/edge", "revision": 1, "resources": []}],
        ),
        (
            "latest/beta",
            [
                {"revision": 1, "resources": [{"name": "boo", "revision": 1}]},
                {"revision": 2, "resources": [{"name": "hoo", "revision": 2}]},
            ],
            [
                {
                    "channel": "latest/beta",
                    "revision": 1,
                    "resources": [{"name": "boo", "revision": 1}],
                },
                {
                    "channel": "latest/beta",
                    "revision": 2,
                    "resources": [{"name": "hoo", "revision": 2}],
                },
            ],
        ),
    ],
)
def test_release_promotion_candidates(
    reusable_store: StoreService, channel, candidates, expected
):
    mock_release = cast(mock.Mock, reusable_store._publisher.release)
    mock_release.reset_mock()

    assert (
        reusable_store.release_promotion_candidates("my-charm", channel, candidates)
        == mock_release.return_value
    )

    mock_release.assert_called_once_with("my-charm", requests=expected)


@pytest.mark.parametrize(
    ("libs", "expected_call"),
    [
        ([], []),
        (
            [CharmLib(lib="my_charm.my_lib", version="1")],
            [{"charm-name": "my-charm", "library-name": "my_lib", "api": 1}],
        ),
        (
            [CharmLib(lib="my_charm.my_lib", version="1.0")],
            [
                {
                    "charm-name": "my-charm",
                    "library-name": "my_lib",
                    "api": 1,
                    "patch": 0,
                }
            ],
        ),
    ],
)
def test_fetch_libraries_metadata(monkeypatch, store, libs, expected_call):
    store.get_libraries_metadata(libs)

    store.anonymous_client.fetch_libraries_metadata.assert_called_once_with(
        expected_call
    )


def test_get_libraries_metadata_name_error(
    monkeypatch, store: StoreService, mock_store_anonymous_client: mock.Mock
) -> None:
    bad_response = requests.Response()
    bad_response.status_code = 400
    bad_response._content = b'{"error-list": [{"code": null, "message": "Items need to include \'library_id\' or \'package_id\'"}]}'
    mock_store_anonymous_client.fetch_libraries_metadata.side_effect = (
        craft_store.errors.StoreServerError(bad_response)
    )

    with pytest.raises(errors.LibraryError, match="One or more declared"):
        store.get_libraries_metadata([CharmLib(lib="boop.snoot", version="-1")])



# Copyright 2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charmcraft-specific build planning service."""

from collections.abc import Iterable
from typing import Any

import craft_platforms
from craft_application.services.buildplan import BuildPlanService


class CharmBuildPlanService(BuildPlanService):
    """A service for generating and filtering build plans."""

    def _gen_exhaustive_build_plan(
        self, project_data: dict[str, Any]
    ) -> Iterable[craft_platforms.BuildInfo]:
        """Generate the exhaustive build plan with craft-platforms.

        :param project_data: The unprocessed project data retrieved from a YAML file.
        :returns: An iterable of BuildInfo objects that make the exhaustive build plan.
        """
        raw_project = self._services.get("project").get_raw()
        if "platforms" in raw_project:
            project_data["platforms"] = raw_project["platforms"]
        else:
            del project_data["platforms"]
        return super()._gen_exhaustive_build_plan(project_data=project_data)



# Copyright 2023-2025 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for package service."""

import pathlib
import sys
import zipfile
from typing import TYPE_CHECKING, Any, cast

import craft_application
import craft_cli.pytest_plugin
import craft_platforms
import distro
import pytest
import pytest_check
from craft_application import util
from craft_platforms import BuildInfo, DebianArchitecture, DistroBase

from charmcraft import const, models, services
from charmcraft.application.main import APP_METADATA

if TYPE_CHECKING:
    from charmcraft.services.package import PackageService

HOST_BASE = craft_platforms.DistroBase.from_linux_distribution(
    distro.LinuxDistribution()
)

SIMPLE_BUILD_BASE = models.charmcraft.Base(
    name=HOST_BASE.distribution, channel=HOST_BASE.series, architectures=["arm64"]
)
SIMPLE_MANIFEST = models.Manifest(
    charmcraft_started_at="2020-03-14T00:00:00+00:00",
    bases=[SIMPLE_BUILD_BASE],
)
MANIFEST_WITH_ATTRIBUTE = models.Manifest.model_validate(
    SIMPLE_MANIFEST.marshal()
    | {
        "analysis": {"attributes": [models.Attribute(name="boop", result="success")]},
    }
)


pytestmark = pytest.mark.skipif(
    sys.platform != "linux", reason="The package service always runs in Linux."
)


@pytest.fixture
def package_service(
    fake_path, simple_charm, service_factory: craft_application.ServiceFactory
):
    fake_project_dir = fake_path / "project"
    fake_project_dir.mkdir(parents=True)

    service_factory.update_kwargs(
        "lifecycle",
        work_dir=fake_path,
        cache_dir=fake_path / "cache",
    )

    return service_factory.get("package")


def test_get_metadata(
    package_service, service_factory: craft_application.ServiceFactory
):
    project = service_factory.get("project").get()
    metadata = models.CharmMetadata(
        name=project.name,
        summary=cast(str, project.summary),
        description=cast(str, project.description),
    )

    assert package_service.metadata == metadata


@pytest.mark.parametrize(
    ("build_plan", "expected_name"),
    [
        pytest.param(
            [
                BuildInfo(
                    platform="distro-1-test64",
                    build_on=DebianArchitecture.RISCV64,
                    build_for=DebianArchitecture.RISCV64,
                    build_base=DistroBase("ubuntu", "24.04"),
                )
            ],
            "example-charm_distro-1-test64.charm",
            id="simple",
        ),
        pytest.param(
            [
                BuildInfo(
                    platform="ubuntu@24.04:riscv64",
                    build_on=DebianArchitecture.RISCV64,
                    build_for=DebianArchitecture.RISCV64,
                    build_base=DistroBase("ubuntu", "24.04"),
                )
            ],
            "example-charm_ubuntu@24.04-riscv64.charm",
            id="multi-base",
        ),
    ],
)
def test_get_charm_name(
    monkeypatch: pytest.MonkeyPatch,
    package_service,
    service_factory: craft_application.ServiceFactory,
    build_plan,
    expected_name,
):
    monkeypatch.setattr(service_factory.get("build_plan"), "plan", lambda: build_plan)

    assert package_service.get_charm_name() == expected_name


@pytest.mark.parametrize(
    ("lint", "attributes"),
    [
        ([], {}),
        (
            [models.CheckResult("lint", "lint", "lint", models.CheckType.LINT, "")],
            {},
        ),
        (
            [models.CheckResult("boop", "success", "", models.CheckType.ATTRIBUTE, "")],
            {
                "analysis": {
                    "attributes": [models.Attribute(name="boop", result="success")]
                },
            },
        ),
    ],
)
def test_get_manifest(
    package_service,
    simple_charm,
    lint,
    attributes,
    service_factory: craft_application.ServiceFactory,
):
    build_item = service_factory.get("build_plan").plan()[0]
    service_factory.get("project").get_platforms()
    expected = models.Manifest.model_validate(
        {
            "charmcraft-started-at": "2020-03-14T00:00:00+00:00",
            "bases": [
                models.charmcraft.Base(
                    name=distro.id(),
                    channel=distro.version(),
                    architectures=[build_item.build_for],
                )
            ],
            **attributes,
        }
    )

    assert package_service.get_manifest(lint) == expected


def test_do_not_overwrite_metadata_yaml(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    package_service,
    service_factory: craft_application.ServiceFactory,
    simple_charm,
):
    dirs = service_factory.get("lifecycle").project_info.dirs
    stage_dir = dirs.stage_dir
    stage_dir.mkdir(exist_ok=True)
    fake_staged_metadata = stage_dir / const.METADATA_FILENAME
    fake_staged_metadata.touch()
    service_factory.get("project").get().parts["reactive"] = {"source": "."}

    package_service.write_metadata(dirs.prime_dir)

    emitter.assert_debug(
        "'metadata.yaml' generated by charm. Not using original project metadata."
    )


def test_do_not_overwrite_actions_yaml(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    package_service,
    service_factory,
    simple_charm,
):
    dirs = service_factory.get("lifecycle").project_info.dirs
    stage_dir = dirs.stage_dir
    stage_dir.mkdir(exist_ok=True)
    fake_staged_metadata = stage_dir / const.JUJU_ACTIONS_FILENAME
    fake_staged_metadata.touch()
    service_factory.get("project").get().parts["reactive"] = {"source": "."}

    package_service.write_metadata(dirs.prime_dir)

    emitter.assert_debug("'actions.yaml' generated by charm. Skipping generation.")


def test_do_not_overwrite_config_yaml(
    emitter: craft_cli.pytest_plugin.RecordingEmitter,
    fake_path,
    package_service,
    service_factory: craft_application.ServiceFactory,
    simple_charm,
):
    dirs = service_factory.get("lifecycle").project_info.dirs
    stage_dir = dirs.stage_dir
    stage_dir.mkdir(exist_ok=True)
    fake_staged_metadata = stage_dir / const.JUJU_CONFIG_FILENAME
    fake_staged_metadata.touch()
    service_factory.get("project").get().parts["reactive"] = {"source": "."}

    package_service.write_metadata(dirs.prime_dir)

    emitter.assert_debug("'config.yaml' generated by charm. Skipping generation.")


# region Tests for getting bases for manifest.yaml
@pytest.mark.parametrize(
    ("bases", "build_item", "expected"),
    [
        (
            [{"name": "ubuntu", "channel": "20.04"}],
            BuildInfo(
                platform=craft_platforms.DebianArchitecture.from_host(),
                build_for=craft_platforms.DebianArchitecture.from_host(),
                build_on=craft_platforms.DebianArchitecture.from_host(),
                build_base=DistroBase("ubuntu", "20.04"),
            ),
            [
                {
                    "name": "ubuntu",
                    "channel": "20.04",
                    "architectures": [craft_platforms.DebianArchitecture.from_host()],
                }
            ],
        ),
        (
            [
                {
                    "build-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["riscv64"],
                        }
                    ],
                    "run-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["all"],
                        },
                    ],
                },
            ],
            BuildInfo(
                platform="riscv64",
                build_for=craft_platforms.DebianArchitecture.RISCV64,
                build_on=craft_platforms.DebianArchitecture.RISCV64,
                build_base=DistroBase("ubuntu", "22.04"),
            ),
            [
                {"name": "ubuntu", "channel": "22.04", "architectures": ["all"]},
            ],
        ),
        pytest.param(
            [
                {
                    "build-on": [{"name": "ubuntu", "channel": "20.04"}],
                    "run-on": [
                        {"name": "ubuntu", "channel": "20.04", "architectures": ["all"]}
                    ],
                },
                {
                    "build-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["amd64", "arm64"],
                        }
                    ],
                    "run-on": [
                        {
                            "name": "ubuntu",
                            "channel": "22.04",
                            "architectures": ["arm64"],
                        }
                    ],
                },
            ],
            BuildInfo(
                platform="amd64",
                build_on=craft_platforms.DebianArchitecture.AMD64,
                build_for=craft_platforms.DebianArchitecture.ARM64,
                build_base=DistroBase("ubuntu", "22.04"),
            ),
            [{"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]}],
            id="cross-compile",
        ),
    ],
)
def test_get_manifest_bases_from_bases(
    monkeypatch: pytest.MonkeyPatch,
    fake_path: pathlib.Path,
    project_path: pathlib.Path,
    bases: list[dict[str, Any]],
    build_item: BuildInfo,
    expected: list[dict[str, Any]],
):
    with (project_path / "charmcraft.yaml").open("w") as f:
        util.dump_yaml(
            {
                "name": "my-charm",
                "description": "",
                "summary": "",
                "type": "charm",
                "bases": bases,
                "parts": {"my-part": {"plugin": "nil"}},
            },
            stream=f,
        )

    services.register_services()
    service_factory = craft_application.ServiceFactory(app=APP_METADATA)
    service_factory.update_kwargs(
        "project",
        project_dir=project_path,
    )
    package_service = cast("PackageService", service_factory.get("package"))

    charm = models.BasesCharm.model_validate(
        {
            "name": "my-charm",
            "description": "",
            "summary": "",
            "type": "charm",
            "bases": bases,
        }
    )
    monkeypatch.setattr(service_factory.get("project"), "get", lambda: charm)
    monkeypatch.setattr(service_factory.get("build_plan"), "plan", lambda: [build_item])

    assert package_service.get_manifest_bases() == [
        models.Base.model_validate(b) for b in expected
    ]


@pytest.mark.parametrize(
    ("base", "build_base", "platforms", "build_item", "expected"),
    [
        pytest.param(
            "ubuntu@24.04",
            None,
            {"test-platform": {"build-on": ["amd64"], "build-for": ["riscv64"]}},
            BuildInfo(
                platform="test-platform",
                build_on=craft_platforms.DebianArchitecture.AMD64,
                build_for=craft_platforms.DebianArchitecture.RISCV64,
                build_base=DistroBase("not-to-be-used", "100"),
            ),
            models.Base(
                # uses the project base
                name="ubuntu",
                channel="24.04",
                architectures=["riscv64"],
            ),
            id="base-from-project",
        ),
        pytest.param(
            "ubuntu@24.04",
            "ubuntu@devel",
            {"test-platform": {"build-on": ["amd64"], "build-for": ["riscv64"]}},
            BuildInfo(
                platform="test-platform",
                build_on=craft_platforms.DebianArchitecture.AMD64,
                build_for=craft_platforms.DebianArchitecture.RISCV64,
                # the BuildInfo will use the build-base, which shouldn't go in the manifest
                build_base=DistroBase("ubuntu", "devel"),
            ),
            models.Base(
                name="ubuntu",
                channel="24.04",
                architectures=["riscv64"],
            ),
            id="ignore-build-base",
        ),
        pytest.param(
            None,
            None,
            {"ubuntu@24.04:amd64": None},
            BuildInfo(
                platform="ubuntu@24.04:amd64",
                build_on=craft_platforms.DebianArchitecture.AMD64,
                build_for=craft_platforms.DebianArchitecture.AMD64,
                build_base=DistroBase("ubuntu", "24.04"),
            ),
            models.Base(
                name="ubuntu",
                channel="24.04",
                architectures=["amd64"],
            ),
            id="multi-base-shorthand",
        ),
        pytest.param(
            None,
            None,
            {
                "test-platform": {
                    "build-on": ["ubuntu@24.04:amd64"],
                    "build-for": ["ubuntu@24.04:riscv64"],
                }
            },
            BuildInfo(
                platform="test-platform",
                build_on=craft_platforms.DebianArchitecture.AMD64,
                build_for=craft_platforms.DebianArchitecture.RISCV64,
                build_base=DistroBase("ubuntu", "24.04"),
            ),
            models.Base(
                name="ubuntu",
                channel="24.04",
                architectures=["riscv64"],
            ),
            id="multi-base-standard",
        ),
    ],
)
def test_get_manifest_bases_from_platforms(
    monkeypatch: pytest.MonkeyPatch,
    project_path: pathlib.Path,
    base,
    build_base,
    platforms,
    build_item,
    expected,
):
    with (project_path / "charmcraft.yaml").open("w") as f:
        util.dump_yaml(
            {
                "name": "my-charm",
                "description": "",
                "summary": "",
                "type": "charm",
                "base": base,
                "build-base": build_base,
                "platforms": platforms,
                "parts": {"my-part": {"plugin": "nil"}},
            },
            stream=f,
        )
    services.register_services()
    service_factory = craft_application.ServiceFactory(app=APP_METADATA)
    service_factory.update_kwargs(
        "project",
        project_dir=project_path,
    )
    package_service = cast("PackageService", service_factory.get("package"))
    service_factory.get("project").configure(platform=None, build_for=None)
    bases = package_service.get_manifest_bases()

    pytest_check.equal(len(bases), 1)
    actual_base = bases[0]
    pytest_check.equal(expected, actual_base)


def test_get_manifest_bases_from_platforms_invalid(
    project_path, package_service, service_factory
):
    with (project_path / "charmcraft.yaml").open("w") as f:
        util.dump_yaml(
            {
                "name": "my-charm",
                "description": "",
                "summary": "",
                "type": "charm",
                "base": None,
                "build-base": None,
                "platforms": {"amd64": None},
                "parts": {"my-part": {"plugin": "nil"}},
            },
            stream=f,
        )
    service_factory._services = {}
    service_factory.get("project").configure(platform=None, build_for=None)

    with pytest.raises(
        craft_platforms.RequiresBaseError,
        match=r"No base or build-base is declared and no base is declared in the platforms section\.",
    ):
        package_service.get_manifest_bases()


# endregion
# region tests for packing the charm
# These tests are modified from test_zipbuild
def test_pack_charm_simple(fake_path, package_service):
    """Build a bunch of files in the zip."""
    build_dir = fake_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "foo.txt"
    testfile1.write_bytes(b"123\x00456")
    subdir = build_dir / "bar"
    subdir.mkdir()
    testfile2 = subdir / "baz.txt"
    testfile2.write_bytes(b"mo\xc3\xb1o")

    charm_path = package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(charm_path)
    assert sorted(x.filename for x in zf.infolist()) == ["bar/baz.txt", "foo.txt"]
    assert zf.read("foo.txt") == b"123\x00456"
    assert zf.read("bar/baz.txt") == b"mo\xc3\xb1o"


def test_zipbuild_symlink_simple(fake_path, package_service):
    """Symlinks are supported."""
    build_dir = fake_path / "somedir"
    build_dir.mkdir()

    testfile1 = build_dir / "real.txt"
    testfile1.write_bytes(b"123\x00456")
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    charm_path = package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(charm_path)
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt", "real.txt"]
    assert zf.read("real.txt") == b"123\x00456"
    assert zf.read("link.txt") == b"123\x00456"


def test_zipbuild_symlink_outside(fake_path, package_service):
    """No matter where the symlink points to."""
    # outside the build dir
    testfile1 = fake_path / "real.txt"
    testfile1.write_bytes(b"123\x00456")

    # inside the build dir
    build_dir = fake_path / "somedir"
    build_dir.mkdir()
    testfile2 = build_dir / "link.txt"
    testfile2.symlink_to(testfile1)

    charm_path = package_service.pack_charm(build_dir, fake_path)

    zf = zipfile.ZipFile(charm_path)
    assert sorted(x.filename for x in zf.infolist()) == ["link.txt"]
    assert zf.read("link.txt") == b"123\x00456"


# endregion



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service class for store interaction."""

from __future__ import annotations

import platform
from collections.abc import Collection, Mapping, Sequence
from typing import Any, cast
from urllib import parse

import craft_application
import craft_store
import distro
from craft_cli import emit
from craft_store import models, publisher
from craft_store.errors import StoreServerError
from overrides import override

from charmcraft import const, env, errors, store
from charmcraft.models import CharmLib
from charmcraft.store import AUTH_DEFAULT_PERMISSIONS, AUTH_DEFAULT_TTL
from charmcraft.store.models import (
    ChannelData,
    Library,
    LibraryMetadataIdRequest,
    LibraryMetadataRequest,
)


class BaseStoreService(craft_application.AppService):
    """Business logic for interacting with the store.

    This service should be easily adjustable for craft-application.
    """

    ClientClass: type[craft_store.StoreClient] = craft_store.StoreClient
    client: craft_store.StoreClient
    _endpoints: craft_store.endpoints.Endpoints = craft_store.endpoints.CHARMHUB
    _environment_auth: str = const.ALTERNATE_AUTH_ENV_VAR

    @property
    def _user_agent(self) -> str:
        """Get the user agent string for the store service."""
        product = self._app.name
        version = self._app.version
        system_information = self._ua_system_info
        return f"{product.title()}/{version} ({system_information})"

    @property
    def _ua_system_info(self) -> str:
        """System information for user agent."""
        segments = [
            f"{platform.system()} {platform.release()}",
            platform.machine(),
            f"{platform.python_implementation()} {platform.python_version()}",
        ]
        if platform.system() == "Linux":
            segments.append(f"{distro.name()} {distro.version()}")

        return "; ".join(segments)

    @property
    def _base_url(self) -> str:
        """Get the store base URL."""
        return env.get_store_config().api_url

    @property
    def _storage_url(self) -> str:
        return env.get_store_config().storage_url

    def setup(self) -> None:
        """Set up the store service."""
        super().setup()

        try:
            self.client = self.ClientClass(
                application_name=self._app.name,
                base_url=self._base_url,
                storage_base_url=self._storage_url,
                endpoints=self._endpoints,
                environment_auth=self._environment_auth,
                user_agent=self._user_agent,
            )
        except craft_store.errors.NoKeyringError:
            emit.progress(
                "WARNING: Cannot get a keyring. Every store interaction that requires "
                "authentication will require you to log in again.",
                permanent=True,
            )
            self.client = self.ClientClass(
                application_name=self._app.name,
                base_url=self._base_url,
                storage_base_url=self._storage_url,
                endpoints=self._endpoints,
                environment_auth=self._environment_auth,
                user_agent=self._user_agent,
                ephemeral=True,
            )

    def _get_description(self, description: str | None = None) -> str:
        """Return the given description or a default one."""
        if description is None:
            return f"{self._app.name}@{platform.node()}"
        return description

    def login(
        self,
        permissions: Sequence[str] = AUTH_DEFAULT_PERMISSIONS,
        description: str | None = None,
        ttl: int = AUTH_DEFAULT_TTL,
        packages: Sequence[craft_store.endpoints.Package] | None = None,
        channels: Sequence[str] | None = None,
    ) -> str:
        """Login to the store."""
        description = self._get_description(description)

        try:
            return self.client.login(
                permissions=permissions,
                description=description,
                ttl=ttl,
                packages=packages,
                channels=channels,
            )
        except craft_store.errors.CredentialsAlreadyAvailable as exc:
            raise errors.CraftError(
                "Cannot login because credentials were found in your system "
                "(which may be no longer valid, though).",
                resolution="Please log out first, then log in again.",
            ) from exc

    def logout(self) -> None:
        """Log out of the store."""
        self.client.logout()

    def get_account_info(self):
        """Get the account details of the logged-in account."""
        return self.client.whoami()["account"]

    def get_credentials(
        self,
        permissions: Sequence[str] = AUTH_DEFAULT_PERMISSIONS,
        description: str | None = None,
        ttl: int = AUTH_DEFAULT_TTL,
        packages: Sequence[craft_store.endpoints.Package] | None = None,
        channels: Sequence[str] | None = None,
    ) -> str:
        """Create a fresh set of login credentials for the store.

        This logs in independent of any credentials currently stored for the application and
        returns the resulting macaroon as a string.
        """
        description = self._get_description(description)

        store = craft_store.StoreClient(
            application_name=self._app.name,
            base_url=self._base_url,
            storage_base_url=self._storage_url,
            endpoints=self._endpoints,
            environment_auth=None,
            user_agent=self._user_agent,
            ephemeral=True,
        )

        return store.login(
            permissions=permissions,
            description=description,
            ttl=ttl,
            packages=packages,
            channels=channels,
        )


class StoreService(BaseStoreService):
    """A Store service specifically for Charmcraft."""

    ClientClass = store.Client
    client: store.Client  # pyright: ignore[reportIncompatibleVariableOverride]
    anonymous_client: store.AnonymousClient
    _publisher: craft_store.PublisherGateway

    @override
    def setup(self) -> None:
        """Set up the store service."""
        super().setup()
        self.anonymous_client = store.AnonymousClient(
            api_base_url=self._base_url,
            storage_base_url=self._storage_url,
        )
        self._auth = craft_store.Auth(
            application_name=self._app.name,
            host=str(parse.urlparse(self._base_url).hostname),
            environment_auth=self._environment_auth,
        )
        self._publisher = craft_store.PublisherGateway(
            base_url=self._base_url,
            namespace="charm",
            auth=self._auth,
        )

    def get_package_metadata(self, name: str) -> publisher.RegisteredName:
        """Get the metadata for a package.

        :param name: The name of the package in this namespace.
        :returns: A RegisteredName model containing store metadata.
        """
        return self._publisher.get_package_metadata(name)

    def release(
        self, name: str, requests: list[publisher.ReleaseRequest]
    ) -> Sequence[publisher.ReleaseResult]:
        """Release one or more revisions to one or more channels.

        :param name: The name of the package to update.
        :param requests: A list of dictionaries containing the requests.
        :returns: A sequence of results of the release requests, as returned
            by the store.

        Each request dictionary requires a "channel" key with the channel name and
        a "revision" key with the revision number. If the revision in the store has
        resources, it requires a "resources" key that is a list of dictionaries
        containing a "name" key with the resource name and a "revision" key with
        the resource number to attach to that channel release.
        """
        return self._publisher.release(name, requests=requests)

    def get_revisions_on_channel(
        self, name: str, channel: str
    ) -> Sequence[Mapping[str, Any]]:
        """Get the current set of revisions on a specific channel.

        :param name: The name on the store to look up.
        :param channel: The channel on which to get the revisions.
        :returns: A sequence of mappings of these, containing their revision,
            bases, resources and version.

        The mapping here may be passed directly into release_promotion_candidates
        in order promote items from one channel to another.
        """
        releases = self._publisher.list_releases(name)
        channel_data = ChannelData.from_str(channel)
        channel_revisions = {
            info.revision: info
            for info in releases.channel_map
            if info.channel == channel_data
        }
        revisions = {
            rev.revision: cast(publisher.CharmRevision, rev)
            for rev in releases.revisions
        }

        return [
            {
                "revision": revision,
                "bases": revisions[revision].bases,
                "resources": [
                    {"name": res.name, "revision": res.revision}
                    for res in info.resources or ()
                ],
                "version": revisions[revision].version,
            }
            for revision, info in channel_revisions.items()
        ]

    def release_promotion_candidates(
        self, name: str, channel: str, candidates: Collection[Mapping[str, Any]]
    ) -> Sequence[publisher.ReleaseResult]:
        """Promote a set of revisions to a specific channel.

        :param name: the store name to operate on.
        :param channel: The channel to which these should be promoted.
        :param candidates: A collection of mappings containing the revision and
            resource revisions to promote.
        :returns: The result of the release in the store.
        """
        requests = [
            publisher.ReleaseRequest(
                channel=channel,
                resources=candidate["resources"],
                revision=candidate["revision"],
            )
            for candidate in candidates
        ]
        return self.release(name, requests)

    def create_tracks(
        self, name: str, *tracks: publisher.CreateTrackRequest
    ) -> Sequence[publisher.Track]:
        """Create tracks in the store.

        :param name: The package name to which the tracks should be attached.
        :param tracks: Each item is a dictionary of the track request.
        :returns: A sequence of the created tracks as dictionaries.
        """
        self._publisher.create_tracks(name, *tracks)
        track_names = {track["name"] for track in tracks}

        return [
            track
            for track in self._publisher.get_package_metadata(name).tracks
            if track.name in track_names
        ]

    def set_resource_revisions_architectures(
        self, name: str, resource_name: str, updates: dict[int, list[str]]
    ) -> Collection[models.resource_revision_model.CharmResourceRevision]:
        """Set the architectures for one or more resource revisions.

        :param name: The name of the charm in the store
        :param resource_name: The name of the specific resource
        :param updates: A mapping of resource revision to its architectures.
        :returns: The updated revisions, as craft_store CharmResourceRevision objects.
        """
        self.client.update_resource_revisions(
            *(
                models.CharmResourceRevisionUpdateRequest(
                    revision=revision,
                    bases=[
                        models.RequestCharmResourceBase(architectures=architectures)
                    ],
                )
                for revision, architectures in updates.items()
            ),
            name=name,
            resource_name=resource_name,
        )
        new_revisions = self.client.list_resource_revisions(
            name=name, resource_name=resource_name
        )
        return [rev for rev in new_revisions if int(rev.revision) in updates]

    def get_libraries_metadata(
        self, libraries: Sequence[CharmLib]
    ) -> Sequence[Library]:
        """Get the metadata for one or more charm libraries.

        :param libraries: A sequence of libraries to request.
        :returns: A sequence of the libraries' metadata in the store.
        """
        store_libs = []
        for lib in libraries:
            charm_name, _, lib_name = lib.lib.partition(".")
            store_lib = LibraryMetadataRequest(
                {
                    "charm-name": charm_name,
                    "library-name": lib_name,
                    "api": lib.api_version,
                }
            )
            if (patch_version := lib.patch_version) is not None:
                store_lib["patch"] = patch_version
            store_libs.append(store_lib)

        try:
            return self.anonymous_client.fetch_libraries_metadata(store_libs)
        except StoreServerError as exc:
            lib_names = [lib.lib for lib in libraries]
            # Type ignore here because error_list is supposed to have string keys, but
            # for whatever reason the store returns a null code for this one.
            # https://bugs.launchpad.net/snapstore-server/+bug/1925065
            if exc.error_list[None]["message"] == (  # type: ignore[index]
                "Items need to include 'library_id' or 'package_id'"
            ):
                raise errors.LibraryError(
                    "One or more declared charm-libs could not be found in the store.",
                    details="Declared charm-libs: " + ", ".join(lib_names),
                    resolution="Check the charm and library names in charmcraft.yaml",
                ) from exc
            raise

    def get_libraries_metadata_by_name(
        self, libraries: Sequence[CharmLib]
    ) -> Mapping[str, Library]:
        """Get a mapping of [charm_name].[library_name] to the requested libraries."""
        return {
            f"{lib.charm_name}.{lib.lib_name}": lib
            for lib in self.get_libraries_metadata(libraries)
        }

    def get_library(
        self,
        charm_name: str,
        *,
        library_id: str,
        api: int | None = None,
        patch: int | None = None,
    ) -> Library:
        """Get a library by charm name and ID from charmhub."""
        return self.anonymous_client.get_library(
            charm_name=charm_name, library_id=library_id, api=api, patch=patch
        )

    def get_libraries_metadata_by_id(self, *lib_id: str) -> Mapping[str, Library]:
        """Get the metadata for a set of libraries by their IDs."""
        store_requests = [
            LibraryMetadataIdRequest({"library-id": lib}) for lib in lib_id
        ]
        return {
            lib.lib_id: lib
            for lib in self.anonymous_client.fetch_libraries_metadata(store_requests)
        }



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Utility functions for craft-parts plugins."""

import pathlib
import shlex
import textwrap
from collections.abc import Collection


def extend_python_build_environment(environment: dict[str, str]) -> dict[str, str]:
    """Extend the build environment for all Python plugins.

    :param environment: the existing environment dictionary
    :returns: the environment dictionary with charmcraft-specific additions.
    """
    return environment | {
        "PARTS_PYTHON_VENV_ARGS": "--without-pip",
    }


def get_charm_copy_commands(
    build_dir: pathlib.Path, install_dir: pathlib.Path
) -> Collection[str]:
    """Get the commands to copy charm source and charmlibs into the install directory.

    The commands will only be included if the relevant directories exist.
    """
    copy_command_base = ["cp", "--archive", "--recursive", "--reflink=auto"]
    src_dir = build_dir / "src"
    libs_dir = build_dir / "lib"

    commands = []
    if src_dir.exists():
        commands.append(
            shlex.join([*copy_command_base, str(src_dir), str(install_dir)])
        )
    if libs_dir.exists():
        commands.append(
            shlex.join([*copy_command_base, str(libs_dir), str(install_dir)])
        )

    return commands


def get_venv_cleanup_commands(venv_path: pathlib.Path, *, keep_bins: bool) -> list[str]:
    """Get a script do Charmcraft-specific venv cleanup.

    :param venv_path: The path to the venv.
    :param keep_bins: Whether to keep the bin directory of the venv.
    :returns: A shell script to do this, as a string.
    """
    venv_bin = venv_path / "bin"
    venv_lib64 = venv_path / "lib64"
    if keep_bins:
        delete_bins = []
    else:
        delete_bins = [
            # Remove all files in venv_bin except `activate`
            "shopt -s extglob",
            f"rm -rf {venv_bin}/!(activate)",
            "shopt -u extglob",
        ]
    update_activate = [
        # Replace hard-coded path in `activate` with portable path
        # "\&" is escape for sed
        'sed -i \'s#^VIRTUAL_ENV=.*$#VIRTUAL_ENV="$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )/.." \\&> /dev/null \\&\\& pwd )"#\' '
        + str(venv_bin / "activate"),
    ]
    delete_lib64 = textwrap.dedent(f"""
        if [ -L '{venv_lib64}' ]; then
          rm -f '{venv_lib64}'
        fi
    """)

    return [*delete_bins, *update_activate, delete_lib64]
